@@ -5,132 +5,26 @@
 
     FLINT is free software: you can redistribute it and/or modify it under
     the terms of the GNU Lesser General Public License (LGPL) as published
-    by the Free Software Foundation; either version 2.1 of the License, or
+    by the Free Software Foundation; either version 3 of the License, or
     (at your option) any later version.  See <https://www.gnu.org/licenses/>.
 */
 
 #ifndef N_POLY_H
 #define N_POLY_H
 
-#ifdef N_POLY_INLINES_C
-#define N_POLY_INLINE
-#else
-#define N_POLY_INLINE static __inline__
-#endif
-
-#include "nmod.h"
+#include "nmod_types.h"
 #include "nmod_vec.h"
 #include "nmod_poly.h"
-#include "fq_nmod.h"
-#include "fq_nmod_poly.h"
+#include "fq_nmod_types.h"
+#include "n_poly_types.h"
 
 #ifdef __cplusplus
- extern "C" {
+extern "C" {
 #endif
 
-/* arrays of ulong */
-typedef struct
-{
-    mp_limb_t * coeffs;
-    slong alloc;
-    slong length;
-} n_poly_struct;
-
-typedef n_poly_struct n_poly_t[1];
-
-typedef n_poly_struct n_fq_poly_struct;
-
-typedef n_poly_t n_fq_poly_t;
-
-/* arrays of arrays of ulong */
-typedef struct
-{
-    n_poly_struct * coeffs;
-    slong alloc;
-    slong length;
-} n_bpoly_struct;
-
-typedef n_bpoly_struct n_bpoly_t[1];
-
-typedef n_bpoly_struct n_fq_bpoly_struct;
-
-typedef n_bpoly_t n_fq_bpoly_t;
-
-/* arrays of arrays of arrays of ulong */
-typedef struct
-{
-    n_bpoly_struct * coeffs;
-    slong alloc;
-    slong length;
-} n_tpoly_struct;
-
-typedef n_tpoly_struct n_tpoly_t[1];
-
-typedef n_tpoly_struct n_fq_tpoly_struct;
-
-typedef n_tpoly_t n_fq_tpoly_t;
-
-/* sparse arrays of ulong */
-typedef struct
-{
-    ulong * exps;
-    mp_limb_t * coeffs;
-    slong length;
-    slong alloc;
-} n_polyu_struct;
-
-typedef n_polyu_struct n_polyu_t[1];
-
-typedef n_polyu_struct n_fq_polyu_struct;
-
-typedef n_polyu_t n_fq_polyu_t;
-
-/*
-    sparse arrays of arrays of ulong
-    n_polyu1n => one exponent is in the exps[i]
-    n_polyu2n => two exponents are packed into the exps[i]
-    ...
-*/
-typedef struct
-{
-    n_poly_struct * coeffs;
-    ulong * exps;
-    slong length;
-    slong alloc;
-} n_polyun_struct;
-
-typedef n_polyun_struct n_polyun_t[1];
-
-typedef n_polyun_struct n_fq_polyun_struct;
-
-typedef n_polyun_t n_fq_polyun_t;
-
-/* n_poly stack */
-typedef struct
-{
-    n_poly_struct ** array;
-    slong alloc;
-    slong top;
-} n_poly_stack_struct;
-
-typedef n_poly_stack_struct n_poly_stack_t[1];
-
-/* n_bpoly stack */
-typedef struct
-{
-    n_bpoly_struct ** array;
-    slong alloc;
-    slong top;
-} n_bpoly_stack_struct;
-
-typedef n_bpoly_stack_struct n_bpoly_stack_t[1];
-
-typedef struct {
-    n_poly_stack_t poly_stack;
-    n_bpoly_stack_t bpoly_stack;
-} n_poly_bpoly_stack_struct;
-
-typedef n_poly_bpoly_stack_struct n_poly_bpoly_stack_t[1];
+#if !defined(FQ_NMOD_H)
+# define fq_nmod_ctx_degree(ctx) ((ctx)->modulus->length - 1)
+#endif
 
 /* n_polyun stack */
 typedef struct
@@ -152,7 +46,7 @@ typedef n_poly_polyun_stack_struct n_poly_polyun_stack_t[1];
 
 /*****************************************************************************/
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_init(n_poly_t A)
 {
     A->length = 0;
@@ -160,7 +54,7 @@ void n_poly_init(n_poly_t A)
     A->coeffs = NULL;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_init2(n_poly_t A, slong alloc)
 {
     A->length = 0;
@@ -170,7 +64,7 @@ void n_poly_init2(n_poly_t A, slong alloc)
         A->coeffs = (mp_limb_t *) flint_malloc(alloc*sizeof(mp_limb_t));
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_clear(n_poly_t A)
 {
     FLINT_ASSERT(A->alloc != 0 || A->coeffs == NULL);
@@ -184,7 +78,7 @@ void n_poly_realloc(n_poly_t A, slong len);
 
 void n_poly_print_pretty(const n_poly_t A, const char * x);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_fit_length(n_poly_t A, slong len)
 {
     if (len > A->alloc)
@@ -192,7 +86,7 @@ void n_poly_fit_length(n_poly_t A, slong len)
 }
 
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void nmod_poly_mock(nmod_poly_t a, const n_poly_t b, nmod_t mod)
 {
     a->coeffs = b->coeffs;
@@ -201,7 +95,7 @@ void nmod_poly_mock(nmod_poly_t a, const n_poly_t b, nmod_t mod)
     a->mod = mod;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mock(n_poly_t a, const nmod_poly_t b)
 {
     a->coeffs = b->coeffs;
@@ -210,7 +104,7 @@ void n_poly_mock(n_poly_t a, const nmod_poly_t b)
 }
 
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_set(n_poly_t A, const n_poly_t B)
 {
     n_poly_fit_length(A, B->length);
@@ -219,42 +113,40 @@ void n_poly_set(n_poly_t A, const n_poly_t B)
 }
 
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_swap(n_poly_t A, n_poly_t B)
 {
-    n_poly_struct t = *B;
-    *B = *A;
-    *A = t;
+    FLINT_SWAP(n_poly_struct, *A, *B);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_poly_normalise(n_poly_t A)
 {
     while (A->length > 0 && A->coeffs[A->length - 1] == 0)
         A->length--;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 slong n_poly_degree(const n_poly_t A)
 {
     FLINT_ASSERT(A->length >= 0);
     return A->length - 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int n_poly_is_one(const n_poly_t A)
 {
     return A->length == 1 && A->coeffs[0] == 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 mp_limb_t n_poly_lead(const n_poly_t A)
 {
     FLINT_ASSERT(A->length > 0);
     return A->coeffs[A->length - 1];
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_one(n_poly_t A)
 {
     n_poly_fit_length(A, 1);
@@ -262,7 +154,7 @@ void n_poly_one(n_poly_t A)
     A->coeffs[0] = 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_set_ui(n_poly_t A, mp_limb_t c)
 {
     n_poly_fit_length(A, 1);
@@ -270,19 +162,19 @@ void n_poly_set_ui(n_poly_t A, mp_limb_t c)
     A->length = (c != 0);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int n_poly_is_zero(const n_poly_t poly)
 {
     return poly->length == 0;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_zero(n_poly_t res)
 {
     res->length = 0;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int n_poly_equal(const n_poly_t a, const n_poly_t b)
 {
     if (a->length != b->length)
@@ -301,7 +193,7 @@ int n_poly_equal(const n_poly_t a, const n_poly_t b)
 
 int n_poly_mod_is_canonical(const n_poly_t A, nmod_t mod);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mod_make_monic(n_poly_t A, const n_poly_t B, nmod_t mod)
 {
     FLINT_ASSERT(B->length > 0);
@@ -310,19 +202,19 @@ void n_poly_mod_make_monic(n_poly_t A, const n_poly_t B, nmod_t mod)
     _nmod_poly_make_monic(A->coeffs, B->coeffs, B->length, mod);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mod_taylor_shift(n_poly_t g, mp_limb_t c, nmod_t mod)
 {
     _nmod_poly_taylor_shift(g->coeffs, c, g->length, mod);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 ulong n_poly_get_coeff(const n_poly_t poly, slong j)
 {
     return (j >= poly->length) ? 0 : poly->coeffs[j];
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_set_coeff_nonzero(n_poly_t A, slong j, ulong c)
 {
     FLINT_ASSERT(c != 0);
@@ -339,7 +231,7 @@ void n_poly_set_coeff(n_poly_t A, slong e, ulong c);
 
 void n_poly_mod_set_coeff_ui(n_poly_t A, slong j, ulong c, nmod_t mod);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_set_nmod_poly(n_poly_t a, const nmod_poly_t b)
 {
     n_poly_fit_length(a, b->length);
@@ -347,7 +239,7 @@ void n_poly_set_nmod_poly(n_poly_t a, const nmod_poly_t b)
     a->length = b->length;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void nmod_poly_set_n_poly(nmod_poly_t a, const n_poly_t b)
 {
     nmod_poly_fit_length(a, b->length);
@@ -355,7 +247,7 @@ void nmod_poly_set_n_poly(nmod_poly_t a, const n_poly_t b)
     a->length = b->length;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_shift_left(n_poly_t A, const n_poly_t B, slong k)
 {
     n_poly_fit_length(A, B->length + k);
@@ -363,7 +255,7 @@ void n_poly_shift_left(n_poly_t A, const n_poly_t B, slong k)
     A->length = B->length + k;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_shift_right(n_poly_t res, const n_poly_t poly, slong k)
 {
     if (k >= poly->length)
@@ -379,7 +271,7 @@ void n_poly_shift_right(n_poly_t res, const n_poly_t poly, slong k)
     }
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_truncate(n_poly_t poly, slong len)
 {
     if (poly->length > len)
@@ -389,7 +281,7 @@ void n_poly_truncate(n_poly_t poly, slong len)
     }
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_poly_mod_scalar_mul_nmod(n_poly_t A, const n_poly_t B, mp_limb_t c,
                                                                     nmod_t mod)
 {
@@ -399,7 +291,7 @@ void _n_poly_mod_scalar_mul_nmod(n_poly_t A, const n_poly_t B, mp_limb_t c,
     A->length = B->length;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_poly_mod_scalar_mul_nmod_inplace(n_poly_t A, mp_limb_t c, nmod_t mod)
 {
     _nmod_vec_scalar_mul_nmod(A->coeffs, A->coeffs, A->length, c, mod);
@@ -411,13 +303,13 @@ void n_poly_mod_scalar_mul_ui(n_poly_t A, const n_poly_t B,
 mp_limb_t n_poly_mod_eval_step2(n_poly_t Acur, const n_poly_t Ainc,
                                                                    nmod_t mod);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 mp_limb_t n_poly_mod_evaluate_nmod(const n_poly_t A, mp_limb_t c, nmod_t mod)
 {
     return _nmod_poly_evaluate_nmod(A->coeffs, A->length, c, mod);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mod_neg(n_poly_t A, const n_poly_t B, nmod_t mod)
 {
     n_poly_fit_length(A, B->length);
@@ -425,7 +317,7 @@ void n_poly_mod_neg(n_poly_t A, const n_poly_t B, nmod_t mod)
     A->length = B->length;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mod_add(n_poly_t A, const n_poly_t B, const n_poly_t C, nmod_t mod)
 {
     slong Alen = FLINT_MAX(B->length, C->length);
@@ -438,7 +330,7 @@ void n_poly_mod_add(n_poly_t A, const n_poly_t B, const n_poly_t C, nmod_t mod)
 void n_poly_mod_add_ui(n_poly_t res, const n_poly_t poly, ulong c, nmod_t ctx);
 
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mod_sub(n_poly_t A, const n_poly_t B, const n_poly_t C, nmod_t mod)
 {
     slong Alen = FLINT_MAX(B->length, C->length);
@@ -448,7 +340,7 @@ void n_poly_mod_sub(n_poly_t A, const n_poly_t B, const n_poly_t C, nmod_t mod)
     _n_poly_normalise(A);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_mod_product_roots_nmod_vec(n_poly_t A, mp_srcptr r, slong n, nmod_t mod)
 {
     n_poly_fit_length(A, n + 1);
@@ -477,7 +369,7 @@ void n_poly_mod_eval2_pow(mp_limb_t * vp, mp_limb_t * vm,
 mp_limb_t n_poly_mod_div_root(n_poly_t Q,
                                     const n_poly_t A, mp_limb_t c, nmod_t ctx);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_poly_mod_mul(n_poly_t A, const n_poly_t B, const n_poly_t C, nmod_t ctx)
 {
     slong Blen = B->length;
@@ -502,7 +394,7 @@ void _n_poly_mod_mul(n_poly_t A, const n_poly_t B, const n_poly_t C, nmod_t ctx)
         _nmod_poly_mul(A->coeffs, C->coeffs, Clen, B->coeffs, Blen, ctx);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_poly_mod_div(n_poly_t Q, const n_poly_t A, const n_poly_t B, nmod_t mod)
 {
     const slong lenA = A->length, lenB = B->length;
@@ -518,7 +410,23 @@ void _n_poly_mod_div(n_poly_t Q, const n_poly_t A, const n_poly_t B, nmod_t mod)
     Q->length = lenA - lenB + 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
+void _n_poly_mod_divexact(n_poly_t Q, const n_poly_t A, const n_poly_t B, nmod_t mod)
+{
+    const slong lenA = A->length, lenB = B->length;
+    FLINT_ASSERT(lenB > 0);
+    FLINT_ASSERT(Q != A && Q != B);
+    if (lenA < lenB)
+    {
+        n_poly_zero(Q);
+        return;
+    }
+    n_poly_fit_length(Q, lenA - lenB + 1);
+    _nmod_poly_divexact(Q->coeffs, A->coeffs, lenA, B->coeffs, lenB, mod);
+    Q->length = lenA - lenB + 1;
+}
+
+FLINT_FORCE_INLINE
 void _n_poly_mod_rem(n_poly_t R, const n_poly_t A, const n_poly_t B, nmod_t mod)
 {
     const slong lenA = A->length, lenB = B->length;
@@ -535,7 +443,7 @@ void _n_poly_mod_rem(n_poly_t R, const n_poly_t A, const n_poly_t B, nmod_t mod)
     _n_poly_normalise(R);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_poly_mod_divrem(n_poly_t Q, n_poly_t R, const n_poly_t A,
                                                   const n_poly_t B, nmod_t mod)
 {
@@ -580,6 +488,9 @@ void n_poly_mod_rem(n_poly_t R, const n_poly_t A, const n_poly_t B,
 void n_poly_mod_divrem(n_poly_t Q, n_poly_t R, const n_poly_t A,
                                                  const n_poly_t B, nmod_t mod);
 
+void n_poly_mod_divexact(n_poly_t Q, const n_poly_t A, const n_poly_t B,
+                                                                   nmod_t mod);
+
 void n_poly_mod_mulmod(n_poly_t res, const n_poly_t poly1,
                            const n_poly_t poly2, const n_poly_t f, nmod_t mod);
 
@@ -605,13 +516,13 @@ void n_poly_mod_mulmod_preinv(n_poly_t A, const n_poly_t B,
 
 /*****************************************************************************/
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 nmod_t fq_nmod_ctx_mod(const fq_nmod_ctx_t ctx)
 {
     return ctx->modulus->mod;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int _n_fq_is_zero(const mp_limb_t * a, slong d)
 {
     do {
@@ -621,7 +532,7 @@ int _n_fq_is_zero(const mp_limb_t * a, slong d)
     return 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_zero(mp_limb_t * a, slong d)
 {
     slong i;
@@ -629,7 +540,7 @@ void _n_fq_zero(mp_limb_t * a, slong d)
         a[i] = 0;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int _n_fq_is_one(const mp_limb_t * a, slong d)
 {
     slong i;
@@ -641,7 +552,7 @@ int _n_fq_is_one(const mp_limb_t * a, slong d)
     return 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int _n_fq_is_ui(const mp_limb_t * a, slong d)
 {
     slong i;
@@ -651,13 +562,13 @@ int _n_fq_is_ui(const mp_limb_t * a, slong d)
     return 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int n_fq_is_one(const mp_limb_t * a, const fq_nmod_ctx_t ctx)
 {
     return _n_fq_is_one(a, fq_nmod_ctx_degree(ctx));
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_one(mp_limb_t * a, slong d)
 {
     slong i;
@@ -666,7 +577,7 @@ void _n_fq_one(mp_limb_t * a, slong d)
         a[i] = 0;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_set_nmod(mp_limb_t * a, mp_limb_t b, slong d)
 {
     slong i;
@@ -679,7 +590,7 @@ void n_fq_gen(
     mp_limb_t * a,
     const fq_nmod_ctx_t ctx);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_set(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -692,7 +603,7 @@ void _n_fq_set(
     } while (i < d);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_swap(
     mp_limb_t * a,      /* length d */
     mp_limb_t * b,      /* length d */
@@ -700,12 +611,12 @@ void _n_fq_swap(
 {
     slong i = 0;
     do {
-        MP_LIMB_SWAP(a[i], b[i]);
+        FLINT_SWAP(mp_limb_t, a[i], b[i]);
         i++;
     } while (i < d);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int _n_fq_equal(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -769,7 +680,7 @@ void n_fq_add_si(
     slong c,
     const fq_nmod_ctx_t ctx);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_fq_add(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -793,7 +704,7 @@ void n_fq_sub_fq_nmod(
     const fq_nmod_t c,
     const fq_nmod_ctx_t ctx);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_fq_sub(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -805,7 +716,7 @@ void n_fq_sub(
     _nmod_vec_sub(a, b, c, d, ctx->modulus->mod);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_add(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -817,7 +728,7 @@ void _n_fq_add(
     _nmod_vec_add(a, b, c, d, mod);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_sub(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -829,7 +740,7 @@ void _n_fq_sub(
     _nmod_vec_sub(a, b, c, d, mod);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_neg(
     mp_limb_t * a,
     const mp_limb_t * b,
@@ -840,18 +751,12 @@ void _n_fq_neg(
     _nmod_vec_neg(a, b, d, mod);
 }
 
-N_POLY_INLINE
 void _n_fq_mul_ui(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
     mp_limb_t c,
     slong d,
-    nmod_t mod)
-{
-    if (c >= mod.n)
-        NMOD_RED(c, c, mod);
-    _nmod_vec_scalar_mul_nmod(a, b, d, c, mod);
-}
+    nmod_t mod);
 
 void _n_fq_madd2(
     mp_limb_t * a,          /* length 2d-1 */
@@ -874,7 +779,7 @@ void _n_fq_reduce(
     mp_limb_t * t);
 
 /* same itch as reduce */
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_reduce2(
     mp_limb_t * a,          /* length d */
     mp_limb_t * b,          /* length 2d-1 */
@@ -893,7 +798,7 @@ void _n_fq_reduce2(
 
 
 #define N_FQ_MUL_ITCH 4
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_mul(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -906,7 +811,7 @@ void _n_fq_mul(
     _n_fq_reduce2(a, t, ctx, t + 2*d);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_addmul(
     mp_limb_t * a,          /* length d */
     const mp_limb_t * b,    /* length d */
@@ -1065,7 +970,7 @@ void n_fq_poly_init2(n_fq_poly_t A, slong alloc, const fq_nmod_ctx_t ctx);
 
 void _n_fq_poly_one(n_fq_poly_t A, slong d);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_fq_poly_one(n_fq_poly_t A, const fq_nmod_ctx_t ctx)
 {
     _n_fq_poly_one(A, fq_nmod_ctx_degree(ctx));
@@ -1077,7 +982,7 @@ int n_fq_poly_is_canonical(
     const n_fq_poly_t a,
     const fq_nmod_ctx_t ctx);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void _n_fq_poly_normalise(n_fq_poly_t A, slong d)
 {
     while (A->length > 0 && _n_fq_is_zero(A->coeffs + d*(A->length - 1), d))
@@ -1270,7 +1175,7 @@ void n_fq_poly_divrem_divconquer_(
     const fq_nmod_ctx_t ctx,
     n_poly_stack_t St);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_fq_poly_divrem_(
     n_poly_t Q,
     n_poly_t R,
@@ -1344,7 +1249,7 @@ void n_fq_poly_eval_pow(
 
 /*****************************************************************************/
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_init(n_bpoly_t A)
 {
     A->coeffs = NULL;
@@ -1354,7 +1259,7 @@ void n_bpoly_init(n_bpoly_t A)
 
 void n_bpoly_clear(n_bpoly_t A);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_swap(n_bpoly_t A, n_bpoly_t B)
 {
     n_bpoly_struct t = *A;
@@ -1365,7 +1270,7 @@ void n_bpoly_swap(n_bpoly_t A, n_bpoly_t B)
 void n_bpoly_print_pretty(const n_bpoly_t A,
                                         const char * xvar, const char * yvar);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_normalise(n_bpoly_t A)
 {
     while (A->length > 0 && n_poly_is_zero(A->coeffs + A->length - 1))
@@ -1374,20 +1279,20 @@ void n_bpoly_normalise(n_bpoly_t A)
 
 void n_bpoly_realloc(n_bpoly_t A, slong len);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_fit_length(n_bpoly_t A, slong len)
 {
     if (len > A->alloc)
         n_bpoly_realloc(A, len);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_zero(n_bpoly_t A)
 {
     A->length = 0;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 int n_bpoly_is_zero(const n_bpoly_t A)
 {
     return A->length == 0;
@@ -1395,7 +1300,7 @@ int n_bpoly_is_zero(const n_bpoly_t A)
 
 void _n_bpoly_set(n_bpoly_t A, const n_bpoly_t B);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_set(n_bpoly_t A, const n_bpoly_t B)
 {
     if (A != B)
@@ -1414,7 +1319,7 @@ void n_bpoly_set_coeff_nonzero(n_bpoly_t A, slong e0, slong e1,
 void n_bpoly_mod_derivative_gen0(n_bpoly_t A, const n_bpoly_t B,
                                                                    nmod_t ctx);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 mp_limb_t n_bpoly_get_coeff(const n_bpoly_t A, slong e0, slong e1)
 {
     if (e0 >= A->length)
@@ -1423,7 +1328,7 @@ mp_limb_t n_bpoly_get_coeff(const n_bpoly_t A, slong e0, slong e1)
         return n_poly_get_coeff(A->coeffs + e0, e1);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 slong n_bpoly_degree0(const n_bpoly_t A)
 {
     return A->length - 1;
@@ -1439,7 +1344,7 @@ void n_bpoly_set_poly_gen0(n_bpoly_t A, const n_poly_t B);
 
 int n_bpoly_mod_is_canonical(const n_bpoly_t A, nmod_t mod);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 ulong n_bpoly_bidegree(const n_bpoly_t A)
 {
     ulong x, y;
@@ -1595,7 +1500,7 @@ int n_fq_bpoly_is_canonical(const n_fq_bpoly_t A, const fq_nmod_ctx_t ctx);
 
 /*****************************************************************************/
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_tpoly_init(n_tpoly_t A)
 {
     A->coeffs = NULL;
@@ -1603,7 +1508,7 @@ void n_tpoly_init(n_tpoly_t A)
     A->length = 0;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_tpoly_swap(n_tpoly_t A, n_tpoly_t B)
 {
     n_tpoly_struct t = *A;
@@ -1617,7 +1522,7 @@ void n_tpoly_clear(n_tpoly_t A);
 
 /*****************************************************************************/
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyu_init(n_polyu_t A)
 {
     A->coeffs = NULL;
@@ -1630,7 +1535,7 @@ void n_polyu_clear(n_polyu_t A);
 
 void n_polyu_realloc(n_polyu_t A, slong len);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyu_fit_length(n_polyu_t A, slong len)
 {
     FLINT_ASSERT(A->alloc >= 0);
@@ -1638,7 +1543,7 @@ void n_polyu_fit_length(n_polyu_t A, slong len)
         n_polyu_realloc(A, len);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyu_swap(n_polyu_t A, n_polyu_t B)
 {
     n_polyu_struct T = *B;
@@ -1709,20 +1614,6 @@ void n_fq_pow_cache_mulpow_fmpz(
 
 /*****************************************************************************/
 
-typedef struct {
-    mp_limb_t * M;
-    mp_limb_t * T;
-    mp_limb_t * Q;
-    mp_limb_t * array;
-    slong alloc;
-    slong d;
-    slong radix;
-    mp_limb_t w;
-} nmod_eval_interp_struct;
-
-typedef nmod_eval_interp_struct nmod_eval_interp_t[1];
-
-
 void nmod_eval_interp_init(nmod_eval_interp_t E);
 
 void nmod_eval_interp_clear(nmod_eval_interp_t E);
@@ -1732,7 +1623,7 @@ int nmod_eval_interp_set_degree_modulus(
     slong deg,
     nmod_t ctx);
 
-N_POLY_INLINE slong nmod_eval_interp_eval_length(nmod_eval_interp_t E)
+FLINT_FORCE_INLINE slong nmod_eval_interp_eval_length(nmod_eval_interp_t E)
 {
     return 1 + E->radix*E->d;
 }
@@ -1762,7 +1653,7 @@ void nmod_eval_interp_from_coeffs_n_fq_poly(
     const fq_nmod_ctx_t ctx);
 
 
-N_POLY_INLINE void nmod_evals_zero(n_poly_t a)
+FLINT_FORCE_INLINE void nmod_evals_zero(n_poly_t a)
 {
     a->length = 0;
 }
@@ -1779,7 +1670,7 @@ void nmod_evals_addmul(n_poly_t a, n_poly_t b, n_poly_t c, slong len,
 void nmod_evals_fmma(n_poly_t a, n_poly_t b, n_poly_t c,
                                 n_poly_t d, n_poly_t e, slong len, nmod_t ctx);
 
-N_POLY_INLINE void n_fq_evals_zero(n_fq_poly_t a)
+FLINT_FORCE_INLINE void n_fq_evals_zero(n_fq_poly_t a)
 {
     a->length = 0;
 }
@@ -1798,7 +1689,7 @@ void n_fq_evals_fmma(n_fq_poly_t a, n_fq_poly_t b, n_fq_poly_t c,
 
 /*****************************************************************************/
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyun_init(n_polyun_t A)
 {
     A->coeffs = NULL;
@@ -1813,7 +1704,7 @@ void n_polyun_clear(n_polyun_t A);
 
 void n_polyun_realloc(n_polyun_t A, slong len);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyun_fit_length(n_polyun_t A, slong len)
 {
     if (len > A->alloc)
@@ -1822,7 +1713,7 @@ void n_polyun_fit_length(n_polyun_t A, slong len)
 
 int n_polyun_mod_is_canonical(const n_polyun_t A, nmod_t mod);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyun_swap(n_polyun_t A, n_polyun_t B)
 {
     n_polyun_struct t = *B;
@@ -1847,7 +1738,7 @@ void n_fq_polyun_set(n_fq_polyun_t A, const n_fq_polyun_t B,
 
 int n_polyun_equal(const n_polyun_t A, const n_polyun_t B);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyun_one(n_polyun_t A)
 {
     n_polyun_fit_length(A, 1);
@@ -1856,7 +1747,7 @@ void n_polyun_one(n_polyun_t A)
     n_poly_one(A->coeffs + 0);
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 ulong n_polyu1n_bidegree(n_polyun_t A)
 {
     ulong x = A->exps[0];
@@ -1912,7 +1803,7 @@ void n_poly_stack_clear(n_poly_stack_t S);
 
 n_poly_struct ** n_poly_stack_fit_request(n_poly_stack_t S, slong k);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 mp_limb_t * n_poly_stack_vec_init(n_poly_stack_t S, slong len)
 {
     n_poly_struct * poly_top;
@@ -1922,14 +1813,14 @@ mp_limb_t * n_poly_stack_vec_init(n_poly_stack_t S, slong len)
     return poly_top->coeffs;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_stack_vec_clear(n_poly_stack_t S)
 {
     FLINT_ASSERT(S->top >= 1);
     S->top -= 1;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 n_poly_struct ** n_poly_stack_request(n_poly_stack_t S, slong k)
 {
     n_poly_struct ** poly_top;
@@ -1938,7 +1829,7 @@ n_poly_struct ** n_poly_stack_request(n_poly_stack_t S, slong k)
     return poly_top;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 n_poly_struct * n_poly_stack_take_top(n_poly_stack_t S)
 {
     /* assume the request for 1 has already been fitted */
@@ -1949,14 +1840,14 @@ n_poly_struct * n_poly_stack_take_top(n_poly_stack_t S)
     return poly_top[0];
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_poly_stack_give_back(n_poly_stack_t S, slong k)
 {
     FLINT_ASSERT(S->top >= k);
     S->top -= k;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 slong n_poly_stack_size(const n_poly_stack_t S)
 {
     return S->top;
@@ -1970,7 +1861,7 @@ void n_bpoly_stack_clear(n_bpoly_stack_t S);
 
 n_bpoly_struct ** n_bpoly_stack_fit_request(n_bpoly_stack_t S, slong k);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 n_bpoly_struct ** n_bpoly_stack_request(n_bpoly_stack_t S, slong k)
 {
     n_bpoly_struct ** bpoly_top;
@@ -1979,7 +1870,7 @@ n_bpoly_struct ** n_bpoly_stack_request(n_bpoly_stack_t S, slong k)
     return bpoly_top;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 n_bpoly_struct * n_bpoly_stack_take_top(n_bpoly_stack_t S)
 {
     /* assume the request for 1 has already been fitted */
@@ -1990,14 +1881,14 @@ n_bpoly_struct * n_bpoly_stack_take_top(n_bpoly_stack_t S)
     return bpoly_top[0];
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_bpoly_stack_give_back(n_bpoly_stack_t S, slong k)
 {
     FLINT_ASSERT(S->top >= k);
     S->top -= k;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 slong n_bpoly_stack_size(const n_bpoly_stack_t S)
 {
     return S->top;
@@ -2011,7 +1902,7 @@ void n_polyun_stack_clear(n_polyun_stack_t S);
 
 n_polyun_struct ** n_polyun_stack_fit_request(n_polyun_stack_t S, slong k);
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 n_polyun_struct ** n_polyun_stack_request(n_polyun_stack_t S, slong k)
 {
     n_polyun_struct ** polyun_top;
@@ -2020,7 +1911,7 @@ n_polyun_struct ** n_polyun_stack_request(n_polyun_stack_t S, slong k)
     return polyun_top;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 n_polyun_struct * n_polyun_stack_take_top(n_polyun_stack_t S)
 {
     /* assume the request for 1 has already been fitted */
@@ -2031,23 +1922,25 @@ n_polyun_struct * n_polyun_stack_take_top(n_polyun_stack_t S)
     return polyun_top[0];
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 void n_polyun_stack_give_back(n_polyun_stack_t S, slong k)
 {
     FLINT_ASSERT(S->top >= k);
     S->top -= k;
 }
 
-N_POLY_INLINE
+FLINT_FORCE_INLINE
 slong n_polyun_stack_size(const n_polyun_stack_t S)
 {
     return S->top;
 }
 
+#if !defined(FQ_NMOD_H)
+# undef fq_nmod_ctx_degree
+#endif
 
 #ifdef __cplusplus
 }
 #endif
 
 #endif
-
